@@ -3,7 +3,7 @@ package HTTP::Any::Curl;
 use strict;
 use warnings;
 
-use Net::Curl::Easy qw(/^CURLOPT_/ CURLE_OK CURLE_ABORTED_BY_CALLBACK CURLINFO_EFFECTIVE_URL);
+use Net::Curl::Easy qw(/^CURLOPT_/ CURLE_OK CURLE_WRITE_ERROR CURLINFO_EFFECTIVE_URL);
 
 
 BEGIN {
@@ -24,7 +24,25 @@ sub do_http {
 		unless ($$opt{headers}{"Content-Type"}) {
 			push @headers, "Content-Type: application/x-www-form-urlencoded";
 		}
-		$easy->setopt(CURLOPT_POSTFIELDS, $$opt{body});
+		if (ref $$opt{body} eq "CODE") {
+			$easy->setopt(CURLOPT_POST, 1);
+			$easy->setopt(CURLOPT_UPLOAD, 1);
+			$easy->setopt(CURLOPT_CUSTOMREQUEST, "POST");
+
+			my $buf = "";
+			my $body_sub = $$opt{body};
+			$easy->setopt(CURLOPT_READFUNCTION, sub {
+				my ( $easy, $maxlen, $uservar ) = @_;
+				$buf ||= $body_sub->();
+				if ($buf) {
+					return \ substr $buf, 0, $maxlen, "";
+				} else {
+					return CURLE_OK;
+				}
+			} );
+		} else {
+			$easy->setopt(CURLOPT_POSTFIELDS, $$opt{body});
+		}
 	}
 
 	$easy->setopt(CURLOPT_HTTPHEADER, \@headers) if @headers;
@@ -45,7 +63,7 @@ sub do_http {
 
 	$easy->setopt(CURLOPT_WRITEHEADER, \ my $headers);
 
-	my $body;
+	my $body = "";
 	$easy->setopt(CURLOPT_FILE, \$body) unless $on_body;
 
 	$easy->setopt(CURLOPT_USERAGENT, $$opt{agent}) if $$opt{agent};
@@ -62,20 +80,16 @@ sub do_http {
 
 	$easy->setopt(CURLOPT_FORBID_REUSE, $$opt{persistent} ? 0 : 1) if exists $$opt{persistent};
 
-	if (my $max_size = $$opt{max_size}) {
-		$easy->setopt(CURLOPT_PROGRESSFUNCTION, sub {
-				my ($easy, $dltotal, $dlnow, $ultotal, $ulnow, $uservar) = @_;
-				if ($dlnow > $max_size) {
-					return 1;
-				}
-				return 0;
-			});
-		$easy->setopt(CURLOPT_NOPROGRESS, 0);
-	}
+	my $max_size = $$opt{max_size};
+	my $aborted_by_max_size = 0;
 
-	if ($on_header or $on_body) {
+	my $body_size = 0;
+
+	if ($max_size or $on_header or $on_body) {
 		my $cb_write = sub {
 			my ($easy, $data, $uservar) = @_;
+			my $size = length $data;
+			$body_size += $size;
 			if ($on_header) {
 				my ($is_success, $headers, $redirects) = headers($easy, $url, $headers);
 				my $r = $on_header->($is_success, $headers, $redirects);
@@ -84,8 +98,14 @@ sub do_http {
 			}
 			if ($on_body) {
 				$on_body->($data) or return 0;
+			} else {
+				$body .= $data;
 			}
-			return length $data;
+			if ($max_size and $body_size > $max_size) {
+				$aborted_by_max_size = 1;
+				return 0;
+			}
+			return $size;
 		};
 		$easy->setopt(CURLOPT_WRITEFUNCTION, $cb_write);
 	}
@@ -95,7 +115,9 @@ sub do_http {
 
 		if ($headers) {
 			my ($is_success, $headers, $redirects) = headers($easy, $url, $headers);
-			$$headers{'client-aborted'} = 'max_size' if $result == CURLE_ABORTED_BY_CALLBACK;
+			if ($result == CURLE_WRITE_ERROR and $aborted_by_max_size) {
+				$$headers{'client-aborted'} = 'max_size';
+			}
 			$cb->($is_success, $body, $headers, $redirects);
 		} else {
 			$cb->(0, undef, { Status => 500, Reason => $easy->error(), URL => $url }, []);
