@@ -106,15 +106,7 @@ sub do_http {
 	my $is_success = $res->is_success || 0;
 
 	if ($$h{'client-aborted'}) {
-		if ($$h{'client-aborted'} eq 'max_size') {
-			$$h{Reason} = 'MaxSize';
-		} elsif ($$h{'client-aborted'} eq 'die') {
-			if ($$h{'x-died'} =~ m/timeout/) {
-				$$h{Reason} = 'Timeout';
-			}
-		}
-		$$h{Status} = 599;
-		$cb->(0, "", $h, \@hr);
+		$cb->($is_success, "", $h, \@hr);
 		return;
 	}
 
